#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the schelchemy project.
#  Please respect the license - more about this in the section (*) below.
#
#  schelchemy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  schelchemy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with schelchemy.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.
from contextlib import contextmanager
from hashlib import md5
from typing import TypeVar

from sqlalchemy import BLOB, Column, String, create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session

from idict.data.compression import pack, unpack

VT = TypeVar("VT")
Base = declarative_base()


class Content(Base):
    __tablename__ = "content"
    id = Column(String(40), primary_key=True)
    blob = Column(BLOB)


def check(key):
    if not isinstance(key, str):
        raise WrongKeyType(f"Key must be string, not {type(key)}.", key)
    return md5(key.encode()).hexdigest() if len(key) not in [32, 40] else key


@contextmanager
def sopen(url="sqlite+pysqlite:///:memory:", autopack=True, ondup="overwrite", deterministic_packing=False, debug=False):
    engine = create_engine(url, echo=debug)
    Base.metadata.create_all(engine)
    with Session(engine, autoflush=False) as session:
        yield ShSQLA(session, autopack, ondup, deterministic_packing)


class ShSQLA:
    r"""
    Dict-like persistence based on SQLAlchemy

    str keys only

    When len(key) not in [32, 40], key is internally hashed to a MD5 hexdigest

    Usage:

    >>> d = ShSQLA("sqlite+pysqlite:////tmp/sqla-test.db")
    >>> d["x"] = 5
    >>> d["x"]
    5
    >>> for k, v in d.items():
    ...     print(k, v)
    9dd4e461268c8034f5c8564e155c67a6 5
    >>> "x" in d
    True
    >>> len(d)
    1
    >>> del d["x"]
    >>> "x" in d
    False
    >>> d
    {}
    >>> with sopen() as db:
    ...     "x" in db
    ...     db
    ...     db["x"] = b"asd"
    ...     db
    ...     db["9dd4e461268c8034f5c8564e155c67a6"] == b"asd"
    ...     "x" in db
    ...     db.x == b"asd"
    ...     del db["x"]
    ...     "x" in db
    False
    {}
    {'9dd4e461268c8034f5c8564e155c67a6': b'asd'}
    True
    True
    True
    False
    """

    def __init__(
            self,
            session="sqlite+pysqlite:///:memory:",
            autopack=True,
            ondup="overwrite",
            deterministic_packing=False,
            debug=False,
    ):
        if isinstance(session, str):

            @contextmanager
            def sessionctx():
                engine = create_engine(url=session, echo=debug)
                Base.metadata.create_all(engine)
                session_ = Session(engine, autoflush=False)
                yield session_
                session_.close()

        else:

            @contextmanager
            def sessionctx():
                yield session

        self.sessionctx = sessionctx
        self.autopack = autopack
        self.ondup = ondup
        self.deterministic_packing = deterministic_packing

    def __contains__(self, key):
        key = check(key)
        with self.sessionctx() as session:
            return session.query(Content).filter_by(id=key).first() is not None

    def __iter__(self):
        with self.sessionctx() as session:
            return (c.id for c in session.query(Content).all())

    def __setitem__(self, key: str, value, packing=True):
        key = check(key)
        if self.autopack and packing:
            value = pack(value, ensure_determinism=self.deterministic_packing)
        elif isinstance(value, str):
            value = value.encode()
        with self.sessionctx() as session:
            if self.ondup == "overwrite":
                session.query(Content).filter_by(id=key).delete()
            if self.ondup == "skip" or session.query(Content).filter_by(id=key).first() is None:
                content = Content(id=key, blob=value)
                session.add(content)
            session.commit()

    def update(self, dic, packing=True):
        with self.sessionctx() as session:
            for k, v in dic.items():
                k = check(k)
                if self.autopack and packing:
                    v = pack(v, ensure_determinism=self.deterministic_packing)
                elif isinstance(v, str):
                    v = v.encode()

                if self.ondup == "overwrite":
                    session.query(Content).filter_by(id=k).delete()
                if self.ondup == "skip" or session.query(Content).filter_by(id=k).first() is None:
                    content = Content(id=k, blob=v)
                    session.add(content)

            session.commit()

    def __getitem__(self, key, packing=True):
        key = check(key)
        with self.sessionctx() as session:
            if ret := session.query(Content).get(key):
                if ret is not None:
                    ret = ret.blob
                    if self.autopack and packing:
                        ret = unpack(ret)
        return ret

    def __delitem__(self, key):
        key = check(key)
        with self.sessionctx() as session:
            session.query(Content).filter_by(id=key).delete()
            session.commit()

    def __getattr__(self, key):
        key_ = check(key)
        if key_ in self:
            return self[key_]
        return self.__getattribute__(key)

    def __len__(self):
        with self.sessionctx() as session:
            return session.query(Content).count()

    def __repr__(self):
        return repr(self.asdict)

    def __str__(self):
        return repr(self)

    @property
    def asdict(self):
        return {k: self[k] for k in self}

    def copy(self):
        raise NotImplementedError

    def keys(self):
        return iter(self)

    def items(self):
        for k in self.keys():
            yield k, self[k]


"""
multiuser                   |   single user
----------------------------------------------
cache = SQLA(session)       |               ->  for already opened/permanent/global sessions
with sqla(url) as cache:    |   shelve      ->  a single session that open/close automatically
cache = SQLA(DBMS url)      |   Disk        ->
----------------------------------------------
Oka(web url); cache         |   dict        -> dict-like cache  
"""


class WrongKeyType(Exception):
    pass



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the idict project.
#  Please respect the license - more about this in the section (*) below.
#
#  idict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  idict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with idict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.
#
import operator
from functools import reduce
from operator import rshift as aop
from operator import xor as cop
from random import Random
from typing import TypeVar, Union, Callable

from garoupa import ø40

import idict.core.frozenidentifieddict as fro
from idict.parameter.ifunctionspace import iFunctionSpace
from idict.parameter.ilet import iLet
from idict.persistence.cache import Cache
from ldict.core.appearance import decolorize
from ldict.core.base import AbstractMutableLazyDict, AbstractLazyDict
from ldict.exception import WrongKeyType

VT = TypeVar("VT")


# TODO: k-fold CV
# TODO (minor):  implement extend, to avoid excessive calculation when batch inserting values
# TODO (minor): let(f, ..., omit=["_name", ...])
# TODO: serialize pandas (and define safe/unsafe unpack), or some of the options bellow
# TODO: Store DF as LazyDF that depends on column metafeatures?
#   or column metafeatures?
#   or depends on a cache? Or both?    LazyDF(cols, indices, cache)
# TODO: document multilevel cache:  .. >> [oka, disk, RAM] >> ..
class Idict(AbstractMutableLazyDict):
    r"""Mutable lazy identified dict for serializable (picklable) pairs str->value

    Usage:

    >>> from idict import idict
    >>> idict().show(colored=False)
    {
        "_id": "0000000000000000000000000000000000000000",
        "_ids": {}
    }
    >>> d = idict(x=5, y=3)
    >>> d.show(colored=False)
    {
        "x": 5,
        "y": 3,
        "_id": "pl_8763f6625970707cdef7dfd31096db4c63c91",
        "_ids": {
            "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd (content: Mj_3bcd9aefb5020343384ae8ccb88fbd872cd8f)",
            "y": "Ku_f738128d6e27a03ec6c2e76d23514b4e998e3 (content: S5_331b7e710abd1443cd82d6b5cdafb9f04d5ab)"
        }
    }
    >>> d["y"]
    3
    >>> d = idict(x=123123, y=88)
    >>> d2 = d >> (lambda x: {"z": x**2})
    >>> d2.ids  # doctest:+ELLIPSIS
    {'z': '...', 'x': 'CF_13e3494690e330a722667f4ab9e067c64022e', 'y': '-j_34f477bff0a6dac76fe447d0e95003e273c62'}
    >>> d2.hosh == d2.identity * d2.ids["z"] * d2.ids["x"] * d2.ids["y"]
    True
    >>> e = d2 >> (lambda x,y: {"w": x/y})
    >>> e.show(colored=False)  # doctest:+ELLIPSIS
    {
        "w": "→(x y)",
        "z": "→(x)",
        "x": 123123,
        "y": 88,
        "_id": "...",
        "_ids": {
            "w": "...",
            "z": "...",
            "x": "CF_13e3494690e330a722667f4ab9e067c64022e (content: I6_bcc6fa41b424962f427be985853b8416b2cff)",
            "y": "-j_34f477bff0a6dac76fe447d0e95003e273c62 (content: 6X_ceb4a9cfad6b3b5a56b49c3484be71dff692a)"
        }
    }
    >>> a = d >> (lambda x: {"z": x**2}) >> (lambda x, y: {"w": x/y})
    >>> b = d >> (lambda x, y: {"w": x/y}) >> (lambda x: {"z": x**2})
    >>> dic = d.asdict  # Converting to dict
    >>> dic
    {'x': 123123, 'y': 88, '_id': 'BZ_2ef6b00691651b6f815ba62b93416aa9b3e80', '_ids': {'x': 'CF_13e3494690e330a722667f4ab9e067c64022e', 'y': '-j_34f477bff0a6dac76fe447d0e95003e273c62'}}
    >>> d2 = idict(dic)  # Reconstructing from a dict
    >>> d2.show(colored=False)
    {
        "x": 123123,
        "y": 88,
        "_id": "BZ_2ef6b00691651b6f815ba62b93416aa9b3e80",
        "_ids": {
            "x": "CF_13e3494690e330a722667f4ab9e067c64022e",
            "y": "-j_34f477bff0a6dac76fe447d0e95003e273c62"
        }
    }
    >>> d == d2
    True
    >>> from idict import Ø
    >>> d = Ø >> {"x": "more content"}
    >>> d.show(colored=False)
    {
        "x": "more content",
        "_id": "zd_6a3f5ec407f6c5c24b89b35ff7cc1793969fd",
        "_ids": {
            "x": "zd_6a3f5ec407f6c5c24b89b35ff7cc1793969fd"
        }
    }
    >>> d = idict() >> {"x": "more content"}
    >>> d.show(colored=False)
    {
        "x": "more content",
        "_id": "zd_6a3f5ec407f6c5c24b89b35ff7cc1793969fd",
        "_ids": {
            "x": "zd_6a3f5ec407f6c5c24b89b35ff7cc1793969fd"
        }
    }
    >>> e.ids.keys()
    dict_keys(['w', 'z', 'x', 'y'])
    >>> del e["z"]
    >>> e.show(colored=False)  # doctest:+ELLIPSIS
    {
        "w": "→(x y)",
        "x": 123123,
        "y": 88,
        "_id": "...",
        "_ids": {
            "w": "...",
            "x": "CF_13e3494690e330a722667f4ab9e067c64022e (content: I6_bcc6fa41b424962f427be985853b8416b2cff)",
            "y": "-j_34f477bff0a6dac76fe447d0e95003e273c62 (content: 6X_ceb4a9cfad6b3b5a56b49c3484be71dff692a)"
        }
    }
    >>> e.hosh == e.identity * e.ids["w"] * e.ids["x"] * e.ids["y"]
    True
    >>> e["x"] = 77
    >>> e.show(colored=False)  # doctest:+ELLIPSIS
    {
        "w": "→(x y)",
        "x": 77,
        "y": 88,
        "_id": "...",
        "_ids": {
            "w": "...",
            "x": "Ye_5b015377273c0cfa49e037788b4c618a9431c (content: 2I_96abdb2113ef4fff69f518ce57969e04f6ded)",
            "y": "-j_34f477bff0a6dac76fe447d0e95003e273c62 (content: 6X_ceb4a9cfad6b3b5a56b49c3484be71dff692a)"
        }
    }
    >>> f = lambda x,y: {"z":x+y}
    >>> d = idict(x=5, y=7)
    >>> d2 = d >> f
    >>> d2.show(colored=False)  # doctest:+ELLIPSIS
    {
        "z": "→(x y)",
        "x": 5,
        "y": 7,
        "_id": "...",
        "_ids": {
            "z": "...",
            "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd (content: Mj_3bcd9aefb5020343384ae8ccb88fbd872cd8f)",
            "y": "WK_6ba95267cec724067d58b3186ecbcaa4253ad (content: 3m_131910d18a892d1b64285250092a4967c8065)"
        }
    }
    >>> c = {}
    >>> d3 = d2 >> [c]
    >>> d3.show(colored=False)  # doctest:+ELLIPSIS
    {
        "z": "→(↑ x y)",
        "x": 5,
        "y": 7,
        "_id": "...",
        "_ids": {
            "z": "...",
            "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd (content: Mj_3bcd9aefb5020343384ae8ccb88fbd872cd8f)",
            "y": "WK_6ba95267cec724067d58b3186ecbcaa4253ad (content: 3m_131910d18a892d1b64285250092a4967c8065)"
        }
    }
    >>> c
    {}
    >>> d3.z
    12
    >>> c  # doctest:+ELLIPSIS
    {'XZJr5b7N0ASkdIo2alN.Crfx7plafTxoqcpJj46e': 12, 'GS_cb0fda15eac732cb08351e71fc359058b93bd': 5, 'WK_6ba95267cec724067d58b3186ecbcaa4253ad': 7, '_NHayCulARZBLYdg-niJJt-SyOeafTxoqcpJj46e': {'_id': 'SNHayCulARZBLYdg-niJJt-SyOeafTxoqcpJj46e', '_ids': {'z': 'XZJr5b7N0ASkdIo2alN.Crfx7plafTxoqcpJj46e', 'x': 'GS_cb0fda15eac732cb08351e71fc359058b93bd', 'y': 'WK_6ba95267cec724067d58b3186ecbcaa4253ad'}}}
    >>> d3.show(colored=False)  # doctest:+ELLIPSIS
    {
        "z": 12,
        "x": 5,
        "y": 7,
        "_id": "...",
        "_ids": {
            "z": "...",
            "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd (content: Mj_3bcd9aefb5020343384ae8ccb88fbd872cd8f)",
            "y": "WK_6ba95267cec724067d58b3186ecbcaa4253ad (content: 3m_131910d18a892d1b64285250092a4967c8065)"
        }
    }
    >>> c = {}
    >>> from idict import setup
    >>> setup(cache=c)
    >>> d3 = d >> f ^ Ø
    >>> d3.show(colored=False)  # doctest:+ELLIPSIS
    {
        "z": "→(↑ x y)",
        "x": 5,
        "y": 7,
        "_id": "...",
        "_ids": {
            "z": "...",
            "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd (content: Mj_3bcd9aefb5020343384ae8ccb88fbd872cd8f)",
            "y": "WK_6ba95267cec724067d58b3186ecbcaa4253ad (content: 3m_131910d18a892d1b64285250092a4967c8065)"
        }
    }
    >>> c
    {}
    >>> d3.z
    12
    >>> c  # doctest:+ELLIPSIS
    {'...': 12, 'GS_cb0fda15eac732cb08351e71fc359058b93bd': 5, 'WK_6ba95267cec724067d58b3186ecbcaa4253ad': 7, '...': {'_id': '...', '_ids': {'z': '...', 'x': 'GS_cb0fda15eac732cb08351e71fc359058b93bd', 'y': 'WK_6ba95267cec724067d58b3186ecbcaa4253ad'}}}
    >>> d3.show(colored=False)  # doctest:+ELLIPSIS
    {
        "z": 12,
        "x": 5,
        "y": 7,
        "_id": "...",
        "_ids": {
            "z": "...",
            "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd (content: Mj_3bcd9aefb5020343384ae8ccb88fbd872cd8f)",
            "y": "WK_6ba95267cec724067d58b3186ecbcaa4253ad (content: 3m_131910d18a892d1b64285250092a4967c8065)"
        }
    }
    >>> f = lambda x: {"y": x ** 2, "_history": ...}
    >>> g = lambda x: {"y":x + 1000, "_history": ...}
    >>> f.metadata = {"id": "b5d6efbc9820dafe0d8fbe87a79adbe9797abc87", "name": "squared", "description": "Some text."}
    >>> g.metadata = {"id": "05d6efbc9820dafe0d8fbe87a79adbe9797abc87", "name": "add1000", "description": "Some text."}
    >>> d = idict(x=3) >> f >> g
    >>> d.show(colored=False)
    {
        "y": "→(x)",
        "_history": "b5d6efbc9820dafe0d8fbe87a79adbe9797abc87 05d6efbc9820dafe0d8fbe87a79adbe9797abc87",
        "x": 3,
        "_id": "FHezbXIcslIpb9mWpbUat7LaEAk1smsieeekmoge",
        "_ids": {
            "y": "Wdha30jmnlZs7fMYWOyoCdQnuoq1smsieeekmoge",
            "_history": "MjnLuTHcrK0ieeQI5ks5ia15W5snEsIJDcm7aDLt",
            "x": "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9 (content: S5_331b7e710abd1443cd82d6b5cdafb9f04d5ab)"
        }
    }
    >>> (idict(x=3).hosh * "b5d6efbc9820dafe0d8fbe87a79adbe9797abc87" * "05d6efbc9820dafe0d8fbe87a79adbe9797abc87").show(colored=False)
    FHezbXIcslIpb9mWpbUat7LaEAk1smsieeekmoge
    >>> a = idict(x=3)
    >>> b = idict(y=5)
    >>> b["d"] = lambda y: a
    >>> cache = {}
    >>> b >>= [cache]
    >>> b.show(colored=False)  # doctest:+ELLIPSIS
    {
        "d": "→(↑ y)",
        "y": 5,
        "_id": "...",
        "_ids": {
            "d": "...",
            "y": "EI_20378979f4669f2e318ae9742e214fd4880d7 (content: Mj_3bcd9aefb5020343384ae8ccb88fbd872cd8f)"
        }
    }
    >>> b.d.show(colored=False)
    {
        "x": 3,
        "_id": "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9",
        "_ids": {
            "x": "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9"
        }
    }
    >>> import json
    >>> print(json.dumps(cache, indent=2))  # doctest:+ELLIPSIS
    {
      "...": {
        "_id": "_E_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9"
      },
      "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9": 3,
      "_E_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9": {
        "_id": "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9",
        "_ids": {
          "x": "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9"
        }
      },
      "EI_20378979f4669f2e318ae9742e214fd4880d7": 5,
      "...": {
        "_id": "...",
        "_ids": {
          "d": "...",
          "y": "EI_20378979f4669f2e318ae9742e214fd4880d7"
        }
      }
    }
    >>> d = idict.fromid("ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9", cache)
    >>> d.show(colored=False)
    {
        "x": "→(↑)",
        "_id": "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9",
        "_ids": {
            "x": "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9"
        }
    }
    >>> d.evaluated.show(colored=False)
    {
        "x": 3,
        "_id": "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9",
        "_ids": {
            "x": "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9"
        }
    }
    >>> idict.fromid(d.id, cache).evaluated.show(colored=False)
    {
        "x": 3,
        "_id": "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9",
        "_ids": {
            "x": "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9"
        }
    }
    >>> e = idict(f=lambda x: 5)
    >>> e.show(colored=False)  # doctest:+ELLIPSIS
    {
        "f": "<function <lambda> at 0x...>",
        "_id": "pqmFNR-9nNXACJsyc6Ro3BA.wfz.mD0zHOgI9vYy",
        "_ids": {
            "f": "pqmFNR-9nNXACJsyc6Ro3BA.wfz.mD0zHOgI9vYy"
        }
    }
    >>> from idict.core.appearance import idict2txt
    >>> d = idict(x=1,y=2)
    >>> decolorize(idict2txt(d, False, False))
    '{\n    "x": 1,\n    "y": 2,\n    "_id": "5G_358b45f49c547174eb4bd687079b30cbbe724",\n    "_ids": "fH_5142f0a4338a1da2ca3159e2d1011981ac890 S-_074b5a806933d64f111a93af359a278402f83"\n}'
    >>> decolorize(idict2txt(d, True, False))
    '{\n    "x": 1,\n    "y": 2,\n    "_id": "5G_358b45f49c547174eb4bd687079b30cbbe724",\n    "_ids": {\n        "x": "fH_5142f0a4338a1da2ca3159e2d1011981ac890 (content: l8_09c7059156c4ed2aea46243e9d4b36c01f272)",\n        "y": "S-_074b5a806933d64f111a93af359a278402f83 (content: -B_305c3d0e44c94a5418d982f7dfe8a537a5c4b)"\n    }\n}'
    >>> cache = {}
    >>> a = idict(x=5) >> (lambda x:{"y": x**x}) >> [[cache]]  # Cache within double brackets has strict/eager behavior.
    >>> a.show(colored=False)  # doctest:+ELLIPSIS
    {
        "y": 3125,
        "x": 5,
        "_id": "...",
        "_ids": {
            "y": "...",
            "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd (content: Mj_3bcd9aefb5020343384ae8ccb88fbd872cd8f)"
        }
    }
    >>> idict(a.id, cache).show(colored=False)  # doctest:+ELLIPSIS
    {
        "y": "→(↑)",
        "x": "→(↑)",
        "_id": "...",
        "_ids": {
            "y": "...",
            "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd (content: Mj_3bcd9aefb5020343384ae8ccb88fbd872cd8f)"
        }
    }
    >>> a.hosh.html
    '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">\n<html>\n<head>\n<meta http-equiv="Content-Type" content="text/html; charset=utf-8">\n<title></title>\n<style type="text/css">\n.ansi2html-content { display: inline; white-space: pre-wrap; word-wrap: break-word; }\n.body_foreground { color: #AAAAAA; }\n.body_background { background-color: #000000; }\n.body_foreground > .bold,.bold > .body_foreground, body.body_foreground > pre > .bold { color: #FFFFFF; font-weight: normal; }\n.inv_foreground { color: #000000; }\n.inv_background { background-color: #AAAAAA; }\n.ansi1 { font-weight: bold; }\n.ansi48-0 { background-color: #000316; }\n.ansi38-117 { color: #87d7ff; }\n.ansi38-153 { color: #afd7ff; }\n.ansi38-189 { color: #d7d7ff; }\n.ansi38-122 { color: #87ffd7; }\n.ansi38-123 { color: #87ffff; }\n.ansi38-159 { color: #afffff; }\n.ansi38-194 { color: #d7ffd7; }\n.ansi38-195 { color: #d7ffff; }\n.ansi38-251 { color: #c6c6c6; }\n</style>\n</head>\n<body class="body_foreground body_background" style="font-size: normal;" >\n<pre class="ansi2html-content">\n<span class="ansi38-123"></span><span class="ansi1 ansi38-123"></span><span class="ansi1 ansi38-123 ansi48-0">q</span><span class="ansi38-159"></span><span class="ansi1 ansi38-159"></span><span class="ansi1 ansi38-159 ansi48-0">1</span><span class="ansi38-195"></span><span class="ansi1 ansi38-195"></span><span class="ansi1 ansi38-195 ansi48-0">O</span><span class="ansi38-195"></span><span class="ansi1 ansi38-195"></span><span class="ansi1 ansi38-195 ansi48-0">Q</span><span class="ansi38-194"></span><span class="ansi1 ansi38-194"></span><span class="ansi1 ansi38-194 ansi48-0">E</span><span class="ansi38-153"></span><span class="ansi1 ansi38-153"></span><span class="ansi1 ansi38-153 ansi48-0">t</span><span class="ansi38-122"></span><span class="ansi1 ansi38-122"></span><span class="ansi1 ansi38-122 ansi48-0">J</span><span class="ansi38-189"></span><span class="ansi1 ansi38-189"></span><span class="ansi1 ansi38-189 ansi48-0">.</span><span class="ansi38-122"></span><span class="ansi1 ansi38-122"></span><span class="ansi1 ansi38-122 ansi48-0">n</span><span class="ansi38-153"></span><span class="ansi1 ansi38-153"></span><span class="ansi1 ansi38-153 ansi48-0">4</span><span class="ansi38-123"></span><span class="ansi1 ansi38-123"></span><span class="ansi1 ansi38-123 ansi48-0">q</span><span class="ansi38-195"></span><span class="ansi1 ansi38-195"></span><span class="ansi1 ansi38-195 ansi48-0">t</span><span class="ansi38-159"></span><span class="ansi1 ansi38-159"></span><span class="ansi1 ansi38-159 ansi48-0">4</span><span class="ansi38-194"></span><span class="ansi1 ansi38-194"></span><span class="ansi1 ansi38-194 ansi48-0">D</span><span class="ansi38-251"></span><span class="ansi1 ansi38-251"></span><span class="ansi1 ansi38-251 ansi48-0">.</span><span class="ansi38-117"></span><span class="ansi1 ansi38-117"></span><span class="ansi1 ansi38-117 ansi48-0">w</span><span class="ansi38-123"></span><span class="ansi1 ansi38-123"></span><span class="ansi1 ansi38-123 ansi48-0">P</span><span class="ansi38-159"></span><span class="ansi1 ansi38-159"></span><span class="ansi1 ansi38-159 ansi48-0">J</span><span class="ansi38-195"></span><span class="ansi1 ansi38-195"></span><span class="ansi1 ansi38-195 ansi48-0">K</span><span class="ansi38-195"></span><span class="ansi1 ansi38-195"></span><span class="ansi1 ansi38-195 ansi48-0">l</span><span class="ansi38-194"></span><span class="ansi1 ansi38-194"></span><span class="ansi1 ansi38-194 ansi48-0">S</span><span class="ansi38-153"></span><span class="ansi1 ansi38-153"></span><span class="ansi1 ansi38-153 ansi48-0">G</span><span class="ansi38-122"></span><span class="ansi1 ansi38-122"></span><span class="ansi1 ansi38-122 ansi48-0">F</span><span class="ansi38-189"></span><span class="ansi1 ansi38-189"></span><span class="ansi1 ansi38-189 ansi48-0">F</span><span class="ansi38-122"></span><span class="ansi1 ansi38-122"></span><span class="ansi1 ansi38-122 ansi48-0">E</span><span class="ansi38-153"></span><span class="ansi1 ansi38-153"></span><span class="ansi1 ansi38-153 ansi48-0">x</span><span class="ansi38-123"></span><span class="ansi1 ansi38-123"></span><span class="ansi1 ansi38-123 ansi48-0">Y</span><span class="ansi38-195"></span><span class="ansi1 ansi38-195"></span><span class="ansi1 ansi38-195 ansi48-0">u</span><span class="ansi38-159"></span><span class="ansi1 ansi38-159"></span><span class="ansi1 ansi38-159 ansi48-0">j</span><span class="ansi38-194"></span><span class="ansi1 ansi38-194"></span><span class="ansi1 ansi38-194 ansi48-0">1</span><span class="ansi38-251"></span><span class="ansi1 ansi38-251"></span><span class="ansi1 ansi38-251 ansi48-0">6</span><span class="ansi38-117"></span><span class="ansi1 ansi38-117"></span><span class="ansi1 ansi38-117 ansi48-0">i</span><span class="ansi38-123"></span><span class="ansi1 ansi38-123"></span><span class="ansi1 ansi38-123 ansi48-0">t</span><span class="ansi38-159"></span><span class="ansi1 ansi38-159"></span><span class="ansi1 ansi38-159 ansi48-0">l</span><span class="ansi38-195"></span><span class="ansi1 ansi38-195"></span><span class="ansi1 ansi38-195 ansi48-0">1</span><span class="ansi38-195"></span><span class="ansi1 ansi38-195"></span><span class="ansi1 ansi38-195 ansi48-0">y</span><span class="ansi38-194"></span><span class="ansi1 ansi38-194"></span><span class="ansi1 ansi38-194 ansi48-0">Q</span><span class="ansi38-153"></span><span class="ansi1 ansi38-153"></span><span class="ansi1 ansi38-153 ansi48-0">B</span><span class="ansi38-122"></span><span class="ansi1 ansi38-122"></span><span class="ansi1 ansi38-122 ansi48-0">L</span><span class="ansi38-189"></span><span class="ansi1 ansi38-189"></span><span class="ansi1 ansi38-189 ansi48-0">4</span>\n</pre>\n</body>\n\n</html>\n'
    >>> a.fields
    ['y', 'x']
    """

    frozen: fro.FrozenIdentifiedDict

    # noinspection PyMissingConstructor
    def __init__(self, /, _dictionary=None, _id=None, _ids=None, rnd=None, identity=ø40, _cloned=None, **kwargs):
        from idict.core.frozenidentifieddict import FrozenIdentifiedDict

        self.identity = identity
        if isinstance(_dictionary, str) and isinstance(_id, (dict, Cache)):
            # Build idict from id+cache.
            if _ids or rnd or _cloned or kwargs:  # pragma: no cover
                raise Exception("Cannot pass more arguments when first argument is id and second argument is cache.")
            # TODO (minor): detect identity (other than ø40) from number of digits
            self.frozen = FrozenIdentifiedDict.fromid(_dictionary, _id, identity=identity)
        elif isinstance(_dictionary, FrozenIdentifiedDict):
            self.frozen = _dictionary
        else:
            self.frozen = FrozenIdentifiedDict(_dictionary, _id, _ids, rnd, identity, _cloned, **kwargs)

    @property
    def id(self):
        return self.hosh.id

    @property
    def ids(self):
        return self.frozen.ids

    @property
    def fields(self):
        return self.frozen.fields

    @property
    def hosh(self):
        return self.frozen.hosh

    @property
    def blobs(self):
        return self.frozen.blobs

    @property
    def hashes(self):
        return self.frozen.hashes

    @property
    def hoshes(self):
        return self.frozen.hoshes

    def __getattr__(self, item):
        if item in self.frozen:
            return self.frozen[item]
        _item = "_" + item
        if _item in self.frozen:
            return self.frozen[_item]
        return self.__getattribute__(item)

    def __delitem__(self, key):
        if not isinstance(key, str):
            raise WrongKeyType(f"Key must be string, not {type(key)}.", key)
        data, blobs, hashes, hoshes = self.data.copy(), self.blobs.copy(), self.hashes.copy(), self.hoshes.copy()
        del data[key]
        for coll in [blobs, hashes, hoshes]:
            if key in coll:
                del coll[key]
        hosh = reduce(operator.mul, [self.identity] + [v for k, v in hoshes.items() if not k.startswith("_")])
        self.frozen = self.frozen.clone(data, _cloned=dict(blobs=blobs, hashes=hashes, hoshes=hoshes, hosh=hosh))

    def clone(self, data=None, rnd=None, _cloned=None):
        return self.frozen.clone(data, rnd, _cloned).asmutable

    def show(self, colored=True, width=None):
        self.frozen.show(colored, width)

    def __rrshift__(self, left: Union[Random, dict, Callable, iFunctionSpace]):
        """
        >>> ({"x": 5} >> Idict(y=2)).show(colored=False)
        {
            "x": 5,
            "y": 2,
            "_id": "xR_15583f654491968d294f9b1622dfb782db241",
            "_ids": {
                "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd (content: Mj_3bcd9aefb5020343384ae8ccb88fbd872cd8f)",
                "y": "S-_074b5a806933d64f111a93af359a278402f83 (content: -B_305c3d0e44c94a5418d982f7dfe8a537a5c4b)"
            }
        }
        >>> from ldict import ldict
        >>> (lambda x: {"y": 5*x}) >> ldict(y = 2)
        «λ × {
            "y": 2
        }»
        """
        if isinstance(left, list) or callable(left):
            return iFunctionSpace(left, aop, self)
        clone = self.__class__(identity=self.identity)
        clone.frozen = left >> self.frozen
        return clone

    def __rshift__(self, other: Union[list, dict, AbstractLazyDict, Callable, iLet, iFunctionSpace, Random]):
        """
        >>> d = Idict(x=2) >> (lambda x: {"y": 2 * x})
        >>> d.ids  # doctest:+ELLIPSIS
        {'y': '...', 'x': 'U8_a7205a343c568c5fe7c4619104ae78bd43279'}
        """
        clone = self.__class__(identity=self.identity)
        clone.frozen = self.frozen >> other
        return clone

    def __rxor__(self, left: Union[Random, dict, Callable, iFunctionSpace]):
        if isinstance(left, list) or callable(left):
            return iFunctionSpace(left, cop, self)
        clone = self.__class__(identity=self.identity)
        clone.frozen = left ^ self.frozen
        return clone

    def __xor__(self, other: Union[dict, AbstractLazyDict, Callable, iLet, iFunctionSpace, Random]):
        clone = self.__class__(identity=self.identity)
        clone.frozen = self.frozen ^ other
        return clone

    @property
    def all(self):
        return self.frozen.all

    @staticmethod
    def fromid(id, cache, identity=ø40):
        from idict.core.frozenidentifieddict import FrozenIdentifiedDict

        return FrozenIdentifiedDict.fromid(id, cache, identity).asmutable

    @staticmethod
    def fromfile(name, output=["df"], output_format="df", include_name=False, identity=ø40):
        from idict.core.frozenidentifieddict import FrozenIdentifiedDict

        return FrozenIdentifiedDict.fromfile(name, output, output_format, include_name, identity).asmutable

    @staticmethod
    def fromtoy(output=["X", "y"], output_format="Xy", identity=ø40):
        from idict.core.frozenidentifieddict import FrozenIdentifiedDict

        return FrozenIdentifiedDict.fromtoy(output, output_format, identity).asmutable

    @staticmethod
    def fromminiarff(output=["df"], output_format="df", identity=ø40):
        from idict.core.frozenidentifieddict import FrozenIdentifiedDict

        return FrozenIdentifiedDict.fromminiarff(output, output_format, identity).asmutable

    @staticmethod
    def fromminicsv(output=["df"], output_format="df", identity=ø40):
        from idict.core.frozenidentifieddict import FrozenIdentifiedDict

        return FrozenIdentifiedDict.fromminicsv(output, output_format, identity).asmutable

    @staticmethod
    def fromopenml(name, version=1, Xout="X", yout="y", identity=ø40):
        from idict.core.frozenidentifieddict import FrozenIdentifiedDict

        return FrozenIdentifiedDict.fromopenml(name, version, Xout, yout, identity).asmutable

    @property
    def metafields(self):
        """
        >>> from idict import idict
        >>> idict(a=1, _b=2, _c=3).metafields
        {'_b': 2, '_c': 3}
        """
        return self.frozen.metafields

    @property
    def trimmed(self):
        """
        >>> from idict import idict
        >>> idict(a=1, _b=2, _history=[1,2,3]).trimmed.show(colored=False)
        {
            "a": 1,
            "_id": "DZ_3d0ee28486d9b4f5e2fe257bc33f1266688fb",
            "_ids": {
                "a": "DZ_3d0ee28486d9b4f5e2fe257bc33f1266688fb"
            }
        }
        """
        return self.frozen.trimmed.asmutable

    def __reduce__(self):
        return self.__class__, ({k: v for k, v in self.data.items()},)

    # def wrapped(self, version, version_id):
    #     """
    #     Wrap a trimmed version of an idict object by a metafield container
    #
    #     The container is identified by `object.id * extra_field__id`.
    #
    #     >>> from idict import idict
    #     >>> d = idict(x=3, _metafield_1=5, _history={"id-of-some-previous-step----------------": 5})
    #     >>> d.show(colored=False)
    #     {
    #         "x": 3,
    #         "_metafield_1": 5,
    #         "_history": "id-of-some-previous-step----------------",
    #         "_id": "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9",
    #         "_ids": {
    #             "x": "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9",
    #             "_metafield_1": "KG_c33bb4404f27e9a7878b29dcb88fbd772cd8f (content: Mj_3bcd9aefb5020343384ae8ccb88fbd872cd8f)",
    #             "_history": "T4_254b4402e8d23be907236a140ff90245285a8 (content: VO_b2b5c7d8ff9718670723c03f0ff9028f085a8)"
    #         }
    #     }
    #     >>> e = d.wrapped("user 1", "uuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuu")
    #     >>> e.show(colored=False)
    #     {
    #         "fields": {
    #             "x": 3,
    #             "_id": "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9",
    #             "_ids": {
    #                 "x": "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9"
    #             }
    #         },
    #         "version": "user 1",
    #         "_metafield_1": 5,
    #         "_history": "id-of-some-previous-step----------------",
    #         "_id": "g.0pzcYOgFTneQ8ojVDcReKkaqDuuuuuuuuuuuuu",
    #         "_ids": {
    #             "fields": "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9",
    #             "version": "uuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuu",
    #             "_metafield_1": "KG_c33bb4404f27e9a7878b29dcb88fbd772cd8f",
    #             "_history": "T4_254b4402e8d23be907236a140ff90245285a8"
    #         }
    #     }
    #     >>> d.hosh == e.fields.hosh
    #     True
    #     """
    #     return self.frozen.wrapped(version, version_id).asmutable



#  Copyright (c) 2022. Davi Pereira dos Santos
#  This file is part of the i-dict project.
#  Please respect the license - more about this in the section (*) below.
#
#  i-dict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  i-dict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with i-dict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#

import shelve
from contextlib import contextmanager
from datetime import datetime, timedelta
from functools import partial
from threading import Thread
from time import sleep

from temporenc import packb, unpackb

from idict.persistence.shelchemy import sopen


def ping(ctx, item, timeout, stop):
    with ctx() as dic:
        while not stop[0]:
            dic[item] = packb(datetime.now())
            t = timeout
            if t is None or t == 0:
                break
            while not stop[0] and t > 0:
                sleep(0.2)
                t -= 0.2


def locker(iterable, dict__url__ctxmgr=None, timeout=None, logstep=1):
    """
    Generator that skips items from 'iterable' already processed before or still being processed

    Item processing is restarted if 'timeout' expires.
    'dict_shelf' is a dict-like or a shelve-like object to store each item status
        when 'None', 'shelve.open("/tmp/locker.db")' will be used
    'logstep' is the frequency of printed messages, 'None' means 'no logs'.
    'timeout'=None keeps the job status as 'started' forever (or until it finishes)

    >>> from time import sleep
    >>> names = ["a","b","c","d","e"]
    >>> storage = {}
    >>> for name in locker(names, dict__url__ctxmgr=storage, timeout=10):
    ...    print(f"Processing {name}")
    ...    sleep(0.1)
    ...    print(f"{name} processed!")
    'a' is new, started
    Processing a
    a processed!
    'a' done
    'b' is new, started
    Processing b
    b processed!
    'b' done
    'c' is new, started
    Processing c
    c processed!
    'c' done
    'd' is new, started
    Processing d
    d processed!
    'd' done
    'e' is new, started
    Processing e
    e processed!
    'e' done
    >>> storage
    {'a': b'd', 'b': b'd', 'c': b'd', 'd': b'd', 'e': b'd'}
    >>> for name in locker(names, dict__url__ctxmgr=storage, timeout=10):
    ...    print(f"Processing {name}")
    ...    sleep(0.1)
    ...    print(f"{name} processed!")
    'a' already done, skipping
    'b' already done, skipping
    'c' already done, skipping
    'd' already done, skipping
    'e' already done, skipping
    """
    if dict__url__ctxmgr is None:
        ctx = partial(shelve.open, "/tmp/locker.db")
    elif isinstance(dict__url__ctxmgr, str):
        ctx = partial(sopen, dict__url__ctxmgr, autopack=False)
    elif isinstance(dict__url__ctxmgr, dict) and hasattr(dict__url__ctxmgr, "__contains__"):
        @contextmanager
        def ctx():
            yield dict__url__ctxmgr
    else:
        ctx = dict__url__ctxmgr

    for c, item in enumerate(iterable):
        with ctx() as dic:
            if item in dic:
                val = dic[item]
                if val == b'd':
                    status, action = 'already done', "skipping"
                elif timeout is not None and datetime.now() > unpackb(val).datetime() + timedelta(seconds=timeout):
                    status, action = "expired", "restarted"
                else:
                    status, action = 'already started', "skipping"
            else:
                status, action = "is new", "started"

        if logstep is not None and c % logstep == 0:
            print(f"'{item}' {status}, {action}")
        if action != "skipping":
            stop = [False]
            t = Thread(target=ping, args=(ctx, item, timeout, stop), daemon=True)
            t.start()
            yield item
            stop[0] = True
            t.join()
            with ctx() as dic:
                dic[item] = b'd'
            if logstep is not None and c % logstep == 0:
                print(f"'{item}' done")



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the idict project.
#  Please respect the license - more about this in the section (*) below.
#
#  idict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  idict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with idict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.

import operator
from functools import reduce
from typing import Dict, Union

from garoupa import Hosh, removal_elem

from idict.core.frozenidentifieddict import FrozenIdentifiedDict
from idict.core.identification import fhosh, blobs_hashes_hoshes
from idict.parameter.ilet import iLet
from ldict.core.base import AbstractLazyDict


def application(self: FrozenIdentifiedDict, other, f, config_hosh, output=None):
    """
    >>> from idict import let
    >>> from garoupa import ø
    >>> d = FrozenIdentifiedDict(x=3)
    >>> f = lambda x: {"y": x**2}
    >>> f.metadata = {"id": "ffffffffffffffffffffffffffffffffffffffff"}
    >>> d2 = application(d, f, f, ø)
    >>> d2.show(colored=False)
    {
        "y": "→(x)",
        "x": 3,
        "_id": "JAsdh.I6zV6p5eiqJkTGxM4g5Dnfffffffffffff",
        "_ids": {
            "y": "23KQ2fUFGjKDTiIseYxUGS9tXqtfffffffffffff",
            "x": "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9 (content: S5_331b7e710abd1443cd82d6b5cdafb9f04d5ab)"
        }
    }
    >>> d2.hosh / f.metadata["id"] == d.id
    True
    """
    f_hosh = f.metadata["id"] if hasattr(f, "metadata") and "id" in f.metadata else fhosh(f, self.identity.version)
    f_hosh_full = self.identity * config_hosh * f_hosh  # d' = d * ħ(config) * f
    if output:
        frozen = self.frozen >> {output: other}
        outputs = [output]
    else:
        frozen = self.frozen >> other
        outputs = frozen.returned
    if "_history" in outputs and ... in frozen.data["_history"]:
        frozen.data["_history"][f.hosh.id] = frozen.data["_history"].pop(...)
    uf = self.hosh * f_hosh_full
    ufu_1 = lambda: solve(self.hoshes, outputs, uf)

    # Reorder items.
    newdata, newhoshes, newblobs, newhashes, = (
        {},
        {},
        self.blobs.copy(),
        self.hashes.copy(),
    )
    noutputs = len(outputs)
    if noutputs == 1:
        k = outputs[0]
        newdata[k] = frozen.data[k]
        newhoshes[k] = ufu_1() if k in self.ids else uf * ~self.hosh
    else:
        ufu_1 = ufu_1()
        acc = self.identity
        c = 0
        last_nonmeta = None
        for k in outputs:
            if not k.startswith("_"):
                last_nonmeta = k
        for i, k in enumerate(outputs):
            newdata[k] = frozen.data[k]
            field_hosh = Hosh(f"{ufu_1.id}-{c}".encode())
            c += 1
            if k == last_nonmeta:
                field_hosh = ~acc * ufu_1
            elif not k.startswith("_"):
                acc *= field_hosh
            newhoshes[k] = field_hosh
            if k in newblobs:
                del newblobs[k]
            if k in newhashes:
                del newhashes[k]
    for k in self.ids:
        if k not in newdata:
            newhoshes[k] = self.hoshes[k]
            newdata[k] = frozen.data[k]

    cloned_internals = dict(blobs=newblobs, hashes=newhashes, hoshes=newhoshes, hosh=uf)
    return self.clone(newdata, _cloned=cloned_internals)


def ihandle_dict(self, dictlike: Union[AbstractLazyDict, dict]):
    """
    >>> from idict.core.frozenidentifieddict import FrozenIdentifiedDict as idict
    >>> d = idict(x=5, y=7, z=8)
    >>> d.show(colored=False)
    {
        "x": 5,
        "y": 7,
        "z": 8,
        "_id": "LB_8c3817cc329ce53b9a02233dbe666a9c05819",
        "_ids": {
            "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd (content: Mj_3bcd9aefb5020343384ae8ccb88fbd872cd8f)",
            "y": "WK_6ba95267cec724067d58b3186ecbcaa4253ad (content: 3m_131910d18a892d1b64285250092a4967c8065)",
            "z": "90_a163ef70987bf1f6157477ae63650fe9161cd (content: fA_de7615cbcc0d4d67bf0d85f2d59addbeccbf8)"
        }
    }
    >>> di = ihandle_dict(d, {"y":None})
    >>> di.show(colored=False)
    {
        "x": 5,
        "y": null,
        "z": 8,
        "_id": "2teR7BgSlcNN0XkG9NShsLJqLZLUp-A7p2gW8E3i",
        "_ids": {
            "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd (content: Mj_3bcd9aefb5020343384ae8ccb88fbd872cd8f)",
            "y": "1hDHzktW1h8pWCu7CsjgxBcPk0FUp-A7p2gW8E3i",
            "z": "90_a163ef70987bf1f6157477ae63650fe9161cd (content: fA_de7615cbcc0d4d67bf0d85f2d59addbeccbf8)"
        }
    }
    >>> di2 = ihandle_dict(di, {"w":lambda x,z: x**z})
    >>> di2.show(colored=False)  # doctest:+ELLIPSIS
    {
        "w": "→(x z)",
        "x": 5,
        "y": null,
        "z": 8,
        "_id": "Q0jOriyrlj63FL3CHO27WUS2rd350InR5ysi4Vfi",
        "_ids": {
            "w": "ZothRZe0R17sOdagBGqPouJ7ZwlcCJOJIvcoXgc0",
            "x": "GS_cb0fda15eac732cb08351e71fc359058b93bd (content: Mj_3bcd9aefb5020343384ae8ccb88fbd872cd8f)",
            "y": "1hDHzktW1h8pWCu7CsjgxBcPk0FUp-A7p2gW8E3i",
            "z": "90_a163ef70987bf1f6157477ae63650fe9161cd (content: fA_de7615cbcc0d4d67bf0d85f2d59addbeccbf8)"
        }
    }
    >>> ihandle_dict(di2, {"x": 55555}).show(colored=False)  # doctest:+ELLIPSIS
    {
        "w": "→(x z)",
        "x": 55555,
        "y": null,
        "z": 8,
        "_id": "fqC1B4bcNgfT9zTqPDTIBKkOfb350InR5ysi4Vfi",
        "_ids": {
            "w": "ZothRZe0R17sOdagBGqPouJ7ZwlcCJOJIvcoXgc0",
            "x": "Xs_cef5ebf024434be670eb110ae10b9067b7f69 (content: 1W_b62d995bf2d19eeb90f0f150cd45cde01a94b)",
            "y": "1hDHzktW1h8pWCu7CsjgxBcPk0FUp-A7p2gW8E3i",
            "z": "90_a163ef70987bf1f6157477ae63650fe9161cd (content: fA_de7615cbcc0d4d67bf0d85f2d59addbeccbf8)"
        }
    }
    >>> (d := ihandle_dict(idict(), {"x": 1555})).show(colored=False)
    {
        "x": 1555,
        "_id": "et_a95995634419ec27251e2f71acf2ef31109d8",
        "_ids": {
            "x": "et_a95995634419ec27251e2f71acf2ef31109d8"
        }
    }
    >>> d >>= lambda x: {"x": x**2}
    >>> d.show(colored=False)  # doctest:+ELLIPSIS
    {
        "x": "→(x)",
        "_id": "...",
        "_ids": {
            "x": "..."
        }
    }
    >>> e = idict(y=7) >> d
    >>> e.show(colored=False)  # doctest:+ELLIPSIS
    {
        "y": 7,
        "x": "→(x)",
        "_id": "kSXVu8FBqNHPSdEaa1dXgyOXI5yWX7zUjhvy-1n0",
        "_ids": {
            "y": "WK_6ba95267cec724067d58b3186ecbcaa4253ad (content: 3m_131910d18a892d1b64285250092a4967c8065)",
            "x": "SSBT4S9CxTb9pCd1US4DCDlXIEwWX7zUjhvy-1n0 (content: cgNWCdJP7XKZASgpBlTRcNAJMPoMv3y0LQcFOoIK)"
        }
    }
    """
    from idict.core.frozenidentifieddict import FrozenIdentifiedDict

    clone = self.clone(rnd=dictlike.rnd) if isinstance(dictlike, AbstractLazyDict) and dictlike.rnd else self.clone()
    for k, v in dictlike.items():
        if v is None:
            clone = delete(clone, k)
        elif k not in ["_id", "_ids"]:
            if isinstance(v, iLet):
                clone = application(clone, v, v.f, v.bytes, k)
            elif callable(v):
                clone = application(clone, v, v, self.identity, k)
            else:
                internals = blobs_hashes_hoshes({k: v}, self.identity, {}, self.identity.version)
                if k in internals["blobs"]:
                    clone.blobs[k] = internals["blobs"][k]
                if k in internals["hashes"]:
                    clone.hashes[k] = internals["hashes"][k]
                clone.hoshes[k] = internals["hoshes"][k]
                hosh = reduce(
                    operator.mul, [self.identity] + [v for k, v in clone.hoshes.items() if not k.startswith("_")]
                )
                internals = dict(blobs=clone.blobs, hashes=clone.hashes, hoshes=clone.hoshes, hosh=hosh)
                del clone.data["_id"]
                del clone.data["_ids"]
                clone = FrozenIdentifiedDict(
                    clone.data, rnd=clone.rnd, identity=self.identity, _cloned=internals, **{k: v}
                )
    return clone


def placeholder(key, f_hosh, identity, hoshes: Dict[str, Hosh]):
    it = iter(hoshes.items())
    while (pair := next(it))[0] != key:
        pass
    # noinspection PyTypeChecker
    oldfield_hosh: Hosh = pair[1]
    right = identity
    for k, v in it:
        right *= v
    field_hosh = oldfield_hosh * right * f_hosh * ~right
    return field_hosh


def solve(hoshes, output, uf: Hosh):
    """
    >>> from idict.core.frozenidentifieddict import FrozenIdentifiedDict as idict
    >>> a = idict(x=3)
    >>> a.show(colored=False)
    {
        "x": 3,
        "_id": "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9",
        "_ids": {
            "x": "ME_bd0a8d9d8158cdbb9d7d4c7af1659ca1dabc9"
        }
    }
    >>> a >>= (lambda x: {"x": x+2})
    >>> a.show(colored=False)
    {
        "x": "→(x)",
        "_id": "y.yeWyCCKqZ36hpE39Y9pta9JiV4XkvCAtYc9OWr",
        "_ids": {
            "x": "y.yeWyCCKqZ36hpE39Y9pta9JiV4XkvCAtYc9OWr"
        }
    }
    >>> a = idict(x=3, y=5) >> (lambda x: {"x": x+2})
    >>> a.hosh == a.hoshes["x"] * a.hoshes["y"]
    True
    >>> a = idict(w=2, x=3) >> (lambda x: {"x": x+2})
    >>> a.hosh == a.hoshes["x"] * a.hoshes["w"]
    True
    >>> a = idict(w=2, x=3, z=1, y=4) >> (lambda x: {"x": x+2})
    >>> a.hosh == a.hoshes["x"] * a.hoshes["w"] * a.hoshes["z"] * a.hoshes["y"]
    True
    >>> a = idict(w=2, x=3, z=1, y=4) >> (lambda w,x,y: {"x": x+2, "a": w*x*y})
    >>> a.hosh == a.hoshes["x"] * a.hoshes["a"] * a.hoshes["w"] * a.hoshes["z"] * a.hoshes["y"]
    True
    >>> a = idict(w=2, x=3, z=1, y=4) >> (lambda w,x,y: {"x": x+2, "y": w*x*y})
    >>> a.hosh == a.hoshes["x"] * a.hoshes["y"] * a.hoshes["w"] * a.hoshes["z"]
    True
    >>> a.show(colored=False)
    {
        "x": "→(w x y)",
        "y": "→(w x y)",
        "w": 2,
        "z": 1,
        "_id": "JN-puhssoMO2iTwz8coREZG5ts9KdJ-3GBbIubB7",
        "_ids": {
            "x": "ypDxvzf3UHQIWNGWB2p5swFgn9eQOGOx5aS5yOxs",
            "y": "oDP6ovudiaAlc2aJ3JoN6tu913zpm9cyAPQBYo3H",
            "w": "Tr_9c39277012db0264069f5c777cdf5ae3c57bc (content: -B_305c3d0e44c94a5418d982f7dfe8a537a5c4b)",
            "z": "fA_f76604de6c618f2740bd6cf44b16673648837 (content: l8_09c7059156c4ed2aea46243e9d4b36c01f272)"
        }
    }
    """
    previous = uf.ø
    for k, v in hoshes.items():
        if k not in output and not k.startswith("_"):
            previous *= v
    return uf * ~previous


def delete(self, k):
    f_hosh = removal_elem(k)
    uf = self.hosh * f_hosh
    newdata = self.data.copy()
    newdata[k] = None
    newhoshes, newblobs, newhashes, = (
        self.hoshes.copy(),
        self.blobs.copy(),
        self.hashes.copy(),
    )
    newhoshes[k] = placeholder(k, f_hosh, self.identity, self.hoshes)
    if k in newblobs:
        del newblobs[k]
    if k in newhashes:
        del newhashes[k]
    return self.clone(newdata, _cloned=dict(blobs=newblobs, hashes=newhashes, hoshes=newhoshes, hosh=uf))



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the i-dict project.
#  Please respect the license - more about this in the section (*) below.
#
#  i-dict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  i-dict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with i-dict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#
from idict.function.dataset import load


def file2df(name):
    if name.endswith(".arff"):
        relation = None
        with open(name) as f:
            for line in f:
                if line[:9].upper() == "@RELATION":
                    relation = line[9:-1]
                    break
        with open(name) as f:
            df = load(f)
        return df, relation or name
    elif name.endswith(".csv"):
        from pandas import read_csv

        return read_csv(name), name
    else:  # pragma: no cover
        raise Exception(f"Unknown extension {name.split('.')[-1]}.")



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the idict project.
#  Please respect the license - more about this in the section (*) below.
#
#  idict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  idict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with idict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.
import time
from abc import abstractmethod

from idict.persistence.cache import Cache


from threading import Thread


def alive(n):
    """Based on https://www.geeksforgeeks.org/start-and-stop-a-thread-in-python/"""
    while n > 0:
        print("contando", n)
        n -= 1
        time.sleep(5)


class CompressedCache(Cache):  # pragma: no cover
    def setblob(self, id, blob):
        # REMINDER: Cannot declare __setitem__ as abstract in this class since SQLA access it from its parent.
        # noinspection PyArgumentList
        self.__setitem__(id, blob, packing=False)

    def getblob(self, id):
        # noinspection PyArgumentList
        return self.__getitem__(id, packing=False)

    def lockid(self, id):
        # TODO (minor):  launch thread at cached (to allow non redundant distributed processing)
        pass
        # t = Thread(target=alive, args=(10,))
        # t.start()

    def unlockid(self, id):
        pass

    @abstractmethod
    def copy(self):
        raise NotImplementedError



from itertools import repeat

from idict.macro import isnumber


def X2histogram(col=0, input="X", output="histogram", bins=8, **kwargs):
    """
    >>> import numpy as np
    >>> from idict import let
    >>> X = np.array([["a", 2.1, 1.6], ["a", 3, 2], ["b", 7, 3]])
    >>> X2histogram(X=X, col=1, bins=2)
    {'histogram': [{'x': '(2.095, 4.55]', 'count': 2}, {'x': '(4.55, 7.0]', 'count': 1}], '_history': Ellipsis}
    >>> from idict import idict
    >>> from idict.function.dataset import df2Xy
    >>> d = idict.fromtoy(output_format="df") >> df2Xy >> X2histogram
    >>> d.histogram
    [{'x': '(0.069, 3.975]', 'count': 11}, {'x': '(3.975, 7.85]', 'count': 5}, {'x': '(7.85, 11.725]', 'count': 3}, {'x': '(11.725, 15.6]', 'count': 0}, {'x': '(15.6, 19.475]', 'count': 0}, {'x': '(19.475, 23.35]', 'count': 0}, {'x': '(23.35, 27.225]', 'count': 0}, {'x': '(27.225, 31.1]', 'count': 1}]
    """
    import numpy as np
    import pandas

    X = kwargs[input]
    vals = X.iloc[:, col] if hasattr(X, "iloc") else X[:, col]
    if isnumber(vals[0]):
        cut = pandas.cut(np.array(list(map(float, vals))), bins, duplicates="drop")
        df = pandas.DataFrame(cut)
        df2 = df.groupby(cut).count()
        dic = df2.to_dict()[0]
    else:
        from pandas import Series

        dic = Series(vals).value_counts()
    result = [{"x": str(k), "count": v} for k, v in dic.items()]
    return {output: result, "_history": ...}


def tofloat(X, k, col):
    if hasattr(X, "iloc"):
        X = X.iloc
    val = X[k, col]
    try:
        return float(val)
    except ValueError:
        return float(list(X[:, col]).index(val))
    except TypeError:
        print(
            f"Warning: Wrong type {type(val)} converted to zero. Look for '?' characters if you provided an ARFF file."
        )
        return 0


def Xy2scatterplot(colx=0, coly=1, Xin="X", yin="y", output="scatterplot", **kwargs):
    """
    >>> import numpy as np
    >>> X = np.array([["c1", 2.1, 1.6], ["c2", 3.2, 2.3], ["c3", 7, 3]])
    >>> y = np.array(["a", "b", "c"])
    >>> Xy2scatterplot(X=X, y=y, colx=1, coly=2)
    {'scatterplot': [{'id': 'a', 'data': [{'x': 2.1, 'y': 1.6}]}, {'id': 'b', 'data': [{'x': 3.2, 'y': 2.3}]}, {'id': 'c', 'data': [{'x': 7.0, 'y': 3.0}]}], '_history': Ellipsis}
    >>> Xy2scatterplot(X=X, y=y, colx=1, coly=0)
    {'scatterplot': [{'id': 'a', 'data': [{'x': 2.1, 'y': 0.0}]}, {'id': 'b', 'data': [{'x': 3.2, 'y': 1.0}]}, {'id': 'c', 'data': [{'x': 7.0, 'y': 2.0}]}], '_history': Ellipsis}
    >>> from idict import idict
    >>> from idict.function.dataset import df2Xy
    >>> d = idict.fromtoy(output_format="df") >> df2Xy >> Xy2scatterplot
    >>> d.scatterplot
    [{'id': '0', 'data': [{'x': 5.1, 'y': 6.4}, {'x': 6.1, 'y': 3.6}, {'x': 3.1, 'y': 2.5}, {'x': 9.1, 'y': 3.5}, {'x': 9.1, 'y': 7.2}, {'x': 7.1, 'y': 6.6}, {'x': 2.1, 'y': 0.1}, {'x': 5.1, 'y': 4.5}, {'x': 1.1, 'y': 3.2}, {'x': 3.1, 'y': 2.5}]}, {'id': '1', 'data': [{'x': 1.1, 'y': 2.5}, {'x': 1.1, 'y': 3.5}, {'x': 4.7, 'y': 4.9}, {'x': 8.3, 'y': 2.9}, {'x': 2.5, 'y': 4.5}, {'x': 0.1, 'y': 4.3}, {'x': 0.1, 'y': 4.0}, {'x': 31.1, 'y': 4.7}, {'x': 2.2, 'y': 8.5}, {'x': 1.1, 'y': 8.5}]}]
    """
    X = kwargs[Xin]
    y = kwargs[yin]
    result = []
    for m in dict(zip(y, repeat(None))):
        inner = []
        for k in range(len(X)):
            left = m if isinstance(m, str) else str(float(m))
            if isinstance(y[k], str):
                right = y[k]
            else:
                right = str(float(y[k]))
            if left == right:
                x_ = tofloat(X, k, colx)
                y_ = tofloat(X, k, coly)
                inner.append({"x": x_, "y": y_})
        result.append({"id": str(m), "data": inner})
    return {output: result, "_history": ...}


X2histogram.metadata = {
    "id": "-----------------------------X2histogram",
    "name": "X2histogram",
    "description": "Generate a histogram for the specified column of a field.",
    "parameters": ...,
    "code": ...,
}
Xy2scatterplot.metadata = {
    "id": "--------------------------Xy2scatterplot",
    "name": "Xy2scatterplot",
    "description": "Generate a scatterplot for the specified two columns of a field.",
    "parameters": ...,
    "code": ...,
}



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the i-dict project.
#  Please respect the license - more about this in the section (*) below.
#
#  i-dict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  i-dict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with i-dict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#
from sklearn.model_selection import StratifiedKFold

from idict import let
from idict.function.wrapper import new, call


def kfold(k=10, seed=0):
    """
    >>> from idict import idict
    >>> d = idict.fromtoy(output_format="Xy")
    >>> d >>= kfold()
    >>> d.indices
    [(array([ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 14, 15, 17, 18,
           19]), array([13, 16])), (array([ 0,  1,  2,  3,  4,  6,  7,  9, 10, 11, 12, 13, 14, 15, 16, 17, 18,
           19]), array([5, 8])), (array([ 1,  2,  3,  4,  5,  6,  8,  9, 10, 11, 12, 13, 14, 15, 16, 17, 18,
           19]), array([0, 7])), (array([ 0,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 15, 16, 17, 18,
           19]), array([ 1, 14])), (array([ 0,  1,  2,  3,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16, 17,
           18]), array([ 4, 19])), (array([ 0,  1,  2,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16, 17,
           19]), array([ 3, 18])), (array([ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 11, 12, 13, 14, 16, 17, 18,
           19]), array([10, 15])), (array([ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 13, 14, 15, 16, 18,
           19]), array([12, 17])), (array([ 0,  1,  3,  4,  5,  6,  7,  8,  9, 10, 12, 13, 14, 15, 16, 17, 18,
           19]), array([ 2, 11])), (array([ 0,  1,  2,  3,  4,  5,  7,  8, 10, 11, 12, 13, 14, 15, 16, 17, 18,
           19]), array([6, 9]))]
    """
    return let(new, class_=StratifiedKFold, config={"n_splits": k, "random_state": seed, "shuffle": True}) >> let(
        call, method="split", input={"X": "X", "y": "y"}, output="indices"
    )



import re


def isplit(source, sep=None, regex=False):
    """
    https://stackoverflow.com/a/9773142/9681577

    generator version of str.split()

    :param source:
        source string (unicode or bytes)

    :param sep:
        separator to split on.

    :param regex:
        if True, will treat sep as regular expression.

    :returns:
        generator yielding elements of string.
    """
    if sep is None:
        # mimic default python behavior
        source = source.strip()
        sep = "\\s+"
        if isinstance(source, bytes):
            sep = sep.encode("ascii")
        regex = True
    if regex:
        # version using re.finditer()
        if not hasattr(sep, "finditer"):
            sep = re.compile(sep)
        start = 0
        for m in sep.finditer(source):
            idx = m.start()
            assert idx >= start
            yield source[start:idx]
            start = m.end()
        yield source[start:]
    else:
        # version using str.find(), less overhead than re.finditer()
        sepsize = len(sep)
        start = 0
        while True:
            idx = source.find(sep, start)
            if idx == -1:
                yield source[start:]
                return
            yield source[start:idx]
            start = idx + sepsize



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the idict project.
#  Please respect the license - more about this in the section (*) below.
#
#  idict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  idict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with idict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.
from pathlib import Path
from typing import TypeVar, Dict, Union

from idict.persistence.disk import Disk

VT = TypeVar("VT")

GLOBAL = {
    "cache": Disk(f"{Path.home()}/{'.idict/shelve.db'}"),
    "compression_cache": {},
    "compression_cachesize": 0,
    "compression_cachelimit": 1_000_000_000,
}


def setup(cache: Union[Disk, Dict[str, VT]] = None, compression_cachelimit_MB: float = None):
    """
    Global behavior of idict

    Parameters
    ----------
    cache
        Dict-like storage accessed through '^' operator.
    compression_cachelimit_MB
        Amount of MBs reserved for keeping compressed values in memory.
        Higher values accelerate persisting original values as compression is already done at hashing.
    """
    if cache is not None:
        GLOBAL["cache"] = cache
    if compression_cachelimit_MB is not None:
        GLOBAL["compression_cachelimit"] = int(compression_cachelimit_MB * 1_000_000)



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the i-dict project.
#  Please respect the license - more about this in the section (*) below.
#
#  i-dict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  i-dict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with i-dict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#
from idict import let
from idict.function.wrapper import call, apply


def df_head(field="df", output="head"):
    """
    >>> from idict import idict
    >>> d = idict.fromtoy(output_format="df")
    >>> d >>= df_head()
    >>> d.head
       attr1  attr2  class
    0    5.1    6.4      0
    1    1.1    2.5      1
    2    6.1    3.6      0
    3    1.1    3.5      1
    4    3.1    2.5      0
    """
    return let(call, field=field, method="head", output=output)


def xywrapper(function, Xin="X", yin="y", Xout="X", yout="y", version=0, **config):
    r"""
    >>> from sklearn.utils import resample
    >>> from idict import idict
    >>> d = idict.fromtoy(output_format="Xy")
    >>> d.y
    array([0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1])
    >>> d >>= xywrapper(resample, n_samples=2, random_state=0)
    >>> d.X
        attr1  attr2
    12    2.1    0.1
    15   31.1    4.7
    >>> d.y
    array([0, 1])
    """
    return let(apply, function=function, input=[Xin, yin], output=[Xout, yout], config=config)



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the idict project.
#  Please respect the license - more about this in the section (*) below.
#
#  idict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  idict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with idict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.


import dis
import pickle
from inspect import signature

from garoupa import Hosh, UT40_4, Identity
from ldict.exception import NoInputException
from orjson import dumps

from idict.data.compression import pack, NondeterminismException

dill_warned = False


def f2bin(f, approach):
    # Add signature.
    fields_and_params = signature(f).parameters.values()
    fields_and_params = {v.name: None if v.default is v.empty else v.default for v in fields_and_params}
    if not fields_and_params:
        raise NoInputException(f"Missing function input parameters.")
    if "_" in fields_and_params:
        return None

    if approach == "clean":
        # Remove line numbers.
        groups = [l for l in dis.Bytecode(f).dis().split("\n\n") if l]
        clean_lines = []
        for group in groups:
            lines = [segment for segment in group.split(" ") if segment][1:]
            clean_lines.append(lines)
        return dumps(clean_lines) + pickle.dumps(fields_and_params, protocol=5)
    if approach == "direct":
        c = f.__code__
        code_bin = c.co_code + str(c.co_consts).encode()
        # TODO (minor): replace pickle for a deterministic dill if possible?
        #  it could allow a broader range of default values (numpy, models)
        return code_bin + pickle.dumps(fields_and_params, protocol=5)
    if approach == "dill":
        # TODO (minor): one advantage of dill here is to be able to hash a custom callable, instead of only functions.
        #  However, preferably, custom callables are expected to provided custom ids.
        global dill_warned
        if not dill_warned:
            dill_warned = True
            print("WARNING: using 'dill' to hash functions is not determinist")
        import dill

        return dill.dumps(f)


def fhosh(f, version, approach="clean"):
    """
    Create hosh with etype="ordered" using bytecode of "f" as binary content for blake3.

    For some insight on the algorithm choice inside GaROUPa, see, e.g.:
    https://news.ycombinator.com/item?id=22021984

    Usage:

    >>> print(fhosh(lambda x: {"z": x**2}, UT40_4))
    p2MGclmVa-FRxu5kFQ65RNjiK42otvusPZ9LGCi4

    >>> print(fhosh(lambda x, name=[1, 2, Ellipsis, ..., 10]: {"z": x**2}, UT40_4))
    3NPAab2SC5lsIz5ekeIQMeQU9EKRW1dYvpUsywyr

    Parameters
    ----------
    f
    version

    Returns
    -------

    """
    if hasattr(f, "hosh"):
        return f.hosh
    if (bin := f2bin(f, approach)) is None:
        f.hosh = Identity(version=version)
    else:
        f.hosh = Hosh(bin, "ordered", version=version)
    return f.hosh


def blobs_hashes_hoshes(data, identity, ids, version):
    """
    >>> from idict import idict
    >>> idict(x=1, y=2, z=3, _ids={"y": "yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy"}).show(colored=False)
    {
        "x": 1,
        "y": 2,
        "z": 3,
        "_id": "Sv8G-WU9SZL90Tus885EWBBf3koyyyyyyyyyyyyy",
        "_ids": {
            "x": "fH_5142f0a4338a1da2ca3159e2d1011981ac890 (content: l8_09c7059156c4ed2aea46243e9d4b36c01f272)",
            "y": "yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy",
            "z": "Nx_e12377018e5ab54023f91f7c6b7aea6676b60 (content: S5_331b7e710abd1443cd82d6b5cdafb9f04d5ab)"
        }
    }
    """
    from idict.core.frozenidentifieddict import FrozenIdentifiedDict
    from idict.core.idict_ import Idict

    blobs = {}
    hashes = {}
    hoshes = {}
    for k, v in data.items():
        if k in ids:
            hoshes[k] = identity * ids[k]
        else:
            if isinstance(v, (Idict, FrozenIdentifiedDict)):
                hashes[k] = v.hosh
            else:
                try:
                    blobs[k] = pack(v)
                    vhosh = identity.h * blobs[k]
                except NondeterminismException:
                    vhosh = fhosh(v, version)
                hashes[k] = vhosh
            try:
                hoshes[k] = hashes[k] ** k.encode()
            except KeyError as e:  # pragma: no cover
                raise Exception(
                    f"{str(e)} is not allowed in field name: {k}. It is only accepted as the first character to indicate a metafield."
                )
    return dict(blobs=blobs, hashes=hashes, hoshes=hoshes)
